from json import dumps,loads
from chastack_bdd.tipos import *
from solteron import Solteron
from sobrecargar import sobrecargar
from secrets import token_urlsafe
from re import findall,match

def formatearValorParaSQL(valor: Any, html : bool = False) -> str:
    """
    Formatea un valor de Python a una representación adecuada para SQL.
    """
    if valor is None:
        return "NULL"
    if isinstance(valor, bool):
        return "1" if valor else "0"
    if isinstance(valor, (int, float)):
        return str(valor)
    if isinstance(valor, (list, tuple)):
        return f"'[{','.join(f"\"{str(v)}\"" for v in valor)}]'"
    if isinstance(valor, Decimal):
        return str(valor.to_eng_string())
    if isinstance(valor, (date, datetime, time)):
        return f"'{valor.isoformat()}'"
    if isinstance(valor, dict):
        return f"'{dumps(valor)}'"
    if isinstance(valor, bytes):
        return f"X'{valor.hex()}'"
    if isinstance(valor, Enum):
        return str(valor.value) if isinstance(valor.value, int) else f"'{valor.name}'"
    if isinstance(valor, str):
        return f"'{valor.replace("'", "''")}'"
        
    return f"'{str(valor).replace("'", "''")}'"

def atributoPublico(nombreAtributo: str) -> str:
    return nombreAtributo.replace('__','',1)

def atributoPrivado(obj: Any, nombreAtributo: str) -> str:
    return f"_{obj.__class__.__name__}__{atributoPublico(nombreAtributo)}"



from chastack_bdd.tipos import *
from chastack_bdd.utiles import *
from chastack_bdd.bdd import ProtocoloBaseDeDatos



class Registro: ...
class Registro:
    __slots__ = (
        '__bdd',
        '__tabla',
        '__id',
    )

    __bdd : ProtocoloBaseDeDatos
    __tabla : str
    __id : int

    @property
    def id(self):
        return self.__id
    
    def __new__(cls, *posicionales,**nominales):
        obj = super(Registro, cls).__new__(cls)
        obj.__tabla = cls.__name__
        return obj        

    @sobrecargar
    def __init__(self, bdd : ProtocoloBaseDeDatos, valores : dict, *, debug : bool =False):
        for atributo in self.__slots__:
            nombre = atributoPublico(atributo)
            valor_SQL : Any = valores.get(nombre,None)
            if valor_SQL is not None:
                valor = valor_SQL
                tipo_esperado : type = self.__class__.__annotations__[atributo]
                if isinstance(valor_SQL,tipo_esperado):
                    valor = valor_SQL
                elif issubclass(tipo_esperado, Decimal):
                    valor : Decimal = Decimal(valor_SQL)
                elif issubclass(tipo_esperado, dict):
                    valor : dict = loads(valor_SQL)
                elif issubclass(tipo_esperado,bool):
                    valor : bool = bool(valor_SQL)
                elif issubclass(tipo_esperado,EnumSQL):
                    valor : tipo_esperado = tipo_esperado.desdeCadena(valor_SQL)
                else:
                    valor = valor_SQL
                setattr(self, atributoPrivado(self,atributo) if '__' in atributo else atributo, valor)
        self.__bdd = bdd
        self.__id = None

    @sobrecargar
    def __init__(self, bdd : ProtocoloBaseDeDatos, id : int, *, debug : bool =False):
        resultado : Resultado
        atributos : tuple[str] = (atributoPublico(atr) for atr in self.__slots__ if atr not in ('__bdd','__tabla'))
        
        with bdd as bdd:
            resultado = bdd\
                        .SELECT(self.tabla,atributos)\
                        .WHERE(id=id)\
                        .ejecutar()\
                        .devolverUnResultado()

        self.__init__(
            bdd,
            resultado
        )
        self.__bdd = bdd
        self.__id = id

    def guardar(self) -> int:
        """Guarda el registro en la tabla correspondiente.
        Si tiene id, se edita un registro existente, 
        de lo contrario se agrega uno nuevo.   

        Devuelve:
        :arg Id int:
            El Id del registro.           

        Levanta:  
        :arg Exception: Propaga errores de la conexión con la BDD  
        :arg Exception: Levanta error si al editar la base con coinciden los id
        """
        match self.__id:
            case None:
                self.__id : int = self.__crear()
            case _: 
                self.__editar()

        return self.__id
    

    def __crear(self) -> int: 
        """Crea un nuevo registro en la tabla correspondiente""" 

        atributos : tuple[str] = (atr for atr in self.__slots__ if '__' not in atr)
        ediciones : dict[str,Any] = {
            atributo : getattr(self,atributo)
            for atributo in atributos
        }
    
        with self.__bdd as bdd:
            id : int = bdd\
                        .UPDATE(self.tabla,**ediciones)\
                        .WHERE(id=self.__id)\
                        .ejecutar()\
                        .devolverIdUltimaInsercion()
            self.__id = id
        
        return self.__id
    
    def __editar(self) -> None: 
        """
        Edita un registro ya existente, dado por el ID, en la tabla correspondiente.
        """


        atributos : tuple[str] = (atr for atr in self.__slots__ if '__' not in atr)
        ediciones : dict[str,Any] = {
            atributo : getattr(self,atributo)
            for atributo in atributos
        }

        with self.__bdd as bdd:
            bdd\
                .UPDATE(self.tabla,**ediciones)\
                .WHERE(id=self.__id)\
                .ejecutar()

    @classmethod
    def devolverRegistros(
        cls,
        bdd : ProtocoloBaseDeDatos,
        *,
        cantidad : Optional[int] = 1000,
        indice : Optional[int] = 0,
        orden : Optional[[dict[str, TipoOrden]]] = {"id":TipoOrden.ASC},
        filtrosJoin : dict[str,str] = None,
        **condiciones) -> tuple[Registro]:

        resultados : tuple[Resultado]
        atributos : tuple[str] = (atributoPublico(atr) for atr in cls.__slots__ if atr not in ('__bdd','__tabla'))
        
        desplazamiento = indice*cantidad 

        bdd\
        .SELECT(cls.__name__, atributos)\
        .WHERE(TipoCondicion.IGUAL,**condiciones)\
        .ORDER_BY(orden)\
        .LIMIT(desplazamiento,cantidad)

        with bdd as bdd:
            resultados = bdd\
                        .ejecutar()\
                        .devolverResultados()
        registros = []
        if resultados:
            for resultado in resultados:
                registros.append(cls(bdd, resultado))

        return tuple(registros)


"""
===============
chastack-bdd
===============
Módulo de Python 3.x con abstracciones útiles para gestión de Bases de Datos y construcción de modelos.

* Repositorio del proyecto: https://github.com/Hernanatn/bdd.py

Derechos de autor (c) 2025 Ch'aska SRL. Distribuído bajo licencia MIT.
Autores:
 - Hernan ATN | herni@cajadeideas.ar 
 - Joaquín C. | correajoaquin7@gmail.com
"""

from chastack_bdd.tipos import *
from chastack_bdd.errores import *
from chastack_bdd.utiles import *
from chastack_bdd.bdd import *
from chastack_bdd.tabla import *
from chastack_bdd.registro import *

if __name__ == '__main__':
    print(__doc__)


import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from chastack_bdd.tipos import *
from chastack_bdd.utiles import *
from chastack_bdd.bdd import ConfigMySQL, BaseDeDatos_MySQL  
from chastack_bdd.tabla import Tabla  
from chastack_bdd.registro import Registro

class Discos(metaclass=Tabla):
    def devolverArtista(self):
        return self.idAutor


config = ConfigMySQL("localhost", "servidor_local", "Servidor!1234", "BaseDePrueba")
bdd = BaseDeDatos_MySQL(config)

disco1 = Discos(bdd=bdd,id=2)


print(disco1.tabla)
print(disco1.id)
print(disco1.tipo.haciaCadena())
print(disco1.soporte)
print(disco1.devolverArtista())

disco2 = Discos(bdd=bdd,id=5)


print(disco2.tabla)
print(disco2.id)
print(disco2.tipo.haciaCadena())
print(disco2.soporte)
print(disco2.devolverArtista())

print(f"{Discos.TipoSoporte.desdeCadena("DIGITAL").value=}")

dds = Discos.devolverRegistros(bdd, cantidad = 25, orden ={"id" : TipoOrden.DESC})
for dd in dds:

    print(dd.tabla)
    print(dd.id)
    print(dd.tipo.haciaCadena())




from chastack_bdd.tipos import *
from chastack_bdd.utiles import *
from chastack_bdd.bdd import ProtocoloBaseDeDatos
from chastack_bdd.registro import Registro

class Tabla(type):
    def __new__(mcs, nombre, bases, atributos):
        if Registro not in bases and nombre != 'Registro':
            bases = (Registro,) + bases
        
        cls = super().__new__(mcs, nombre, bases, atributos)
        
        cls.__tabla = nombre
        setattr(cls, "tabla", property(lambda cls : cls.__tabla))

        
        if not hasattr(cls, '__annotations__'):
            cls.__annotations__ = {}
        
        return cls

    def __init__(cls, nombre, ancestros, diccionario):
        cls.__INICIALIZADA = False
        cls.__DEBUG = False


    def __call__(cls, bdd: ProtocoloBaseDeDatos, *posicionales, **nominales): 
        if nominales and nominales.get("debug", False):
            cls.__DEBUG = True
        debug = lambda msj: print(f"[DEBUG] {msj}") if cls.__DEBUG else lambda msj: None
        debug(f"{posicionales}, {nominales}")
        debug(f"Se llamó a la clase {cls}. Instanciando objeto.")
        debug(f"{cls} {'ya' if cls.__INICIALIZADA else 'no'} estaba inicializada.")
        if not cls.__INICIALIZADA: 
            slots :list[str] = []        
            anotaciones : dict[str,type] = {}
            with bdd as bdd:
                resultados = bdd.DESCRIBE(cls.__tabla).ejecutar().devolverResultados()
                
                debug(f"Inicializando modelo para: {cls.__tabla}.")
                for columna in resultados:
                    nombre_campo = columna.get('Field')
                    es_clave = columna.get('Key') == "PRI"
                    es_auto = "auto_increment" in columna.get("Extra", "").lower() or "default_generated" in columna.get("Extra", "").lower() or "auto_generated" in columna.get("Extra", "").lower()
                    
                    nombre_attr = f"__{nombre_campo}" if es_clave or es_auto else nombre_campo
                    
                    tipo = cls.__resolverTipo(columna.get('Type'), nombre_campo)
                    debug(f"| {str(nombre_campo):<40} | {str(columna.get('Type','')):<40} | {str(tipo):<40}")
                    
                    if nombre_attr not in cls.__slots__:
                        slots.append(nombre_attr)
                    anotaciones.update({
                        nombre_attr : tipo
                    })
                    
                    
                    if es_clave:
                        setattr(cls, nombre_campo, property(lambda self, name=nombre_campo: getattr(self, atributoPrivado(self,name))))
        
            cls.__slots__ = cls.__slots__ + tuple(slots)
            cls.__annotations__.update(anotaciones)
            cls.__INICIALIZADA = True
        
        instancia = super().__call__(bdd, *posicionales, **nominales)
        setattr(instancia, atributoPrivado(instancia,"__bdd"),bdd)
        return instancia
    
    @classmethod
    def __resolverTipo(cls, tipo_sql: str, nombre_columna: Optional[str]) -> type:
        """
        Deduce y devuelve un tipo de Python en base al tipo declarado en MySQL para la columna.
        Si encuentra un ENUM, crea un enum de Python y lo guarda como una constante de la clase.
        
        Parámetros:
            :arg tipo_sql str: El tipo definido en MySQL
            :arg nombre_columna Optional[str]: El nombre de la columna (útil para enums)
        
        Devuelve:
            :arg tipo `type`: el tipo python correspondiente (o `Any`)
        """
        tipo_declarado: Optional[Match[AnyStr]] = match(r'([a-z]+)(\(.*\))?', tipo_sql.lower())
        if not tipo_declarado:
            return Any

        tipo_base: str = tipo_declarado.group(1)
        parametros: str = tipo_declarado.group(2) if tipo_declarado.group(2) else ""
        tipo_completo: str = tipo_base + parametros

        tipos: dict[str, type] = {
            'tinyint': int,
            'smallint': int,
            'mediumint': int,
            'int': int,
            'bigint': int,
            'float': float,
            'double': float,
            'decimal': Decimal,
            'datetime': datetime,
            'timestamp': datetime,
            'date': date,
            'time': time,
            'char': str,
            'varchar': str,
            'text': str,
            'mediumtext': str,
            'longtext': str,
            'tinytext': str,
            'boolean': bool,
            'bool': bool,
            'tinyint(1)': bool,
            'blob': bytearray,
            'mediumblob': bytearray,
            'longblob': bytearray,
            'tinyblob': bytearray,
            'binary': bytearray,
            'varbinary': bytearray,
            'json': dict,
        }

        
        if tipo_base == 'enum':
            valores_enum: list[Any] = findall(r"'([^']*)'", tipo_sql)
            dicc_enum: dict[str, int] = {'_invalido': 0}
            for i, val in enumerate(valores_enum, 1):   
                dicc_enum[val] = i
            
            nombre_enum: str = f"Tipo{nombre_columna.capitalize()}" if nombre_columna else f"__ENUM_{token_urlsafe(4)}"
            clase_enum: type = type(
                nombre_enum,
                (EnumSQL, Enum),
                dicc_enum
            )
            
            setattr(cls,nombre_enum,clase_enum)
            return clase_enum

        
        if tipo_completo in tipos:
            return tipos[tipo_completo]
        return tipos.get(tipo_base, Any)


from enum import Enum, EnumType as _EnumMeta, EnumDict as _EnumDict

class EnumSQLMeta(_EnumMeta):
    """
    Metaclase para controlar las reglas personalizadas sobre las enumeraciones.
    Asegura que `_invalido` siempre tenga el valor 0 y ningún otro miembro lo tenga.
    """
    @classmethod
    def __prepare__(metacls, cls, bases, **kwds):
        """Copiado de enum.EnumType con la modificación de que esta metaclase no inhibe la herencia"""
        #metacls._check_for_existing_members_(cls, bases)
        enum_dict = _EnumDict()
        enum_dict._cls_name = cls
        member_type, first_enum = metacls._get_mixins_(cls, bases)
        if first_enum is not None:
            enum_dict['_generate_next_value_'] = getattr(
                    first_enum, '_generate_next_value_', None,
                    )
        return enum_dict
    def __new__(cls, nombre, bases, diccionario):
        diccionario_enum = EnumSQLMeta.__prepare__(nombre,bases,**diccionario)
        diccionario_enum.update(diccionario)
        nueva_clase = _EnumMeta.__new__(cls, nombre, bases, diccionario_enum)

        # Verificar que la clase tiene '_invalido' con valor 0
        if '_invalido' not in nueva_clase.__members__: ...
            #raise ValueError(f"La enumeración {nombre} debe tener un miembro '_invalido' con valor 0.")
        
        # Verificar que ningún otro miembro tenga valor 0.
        for miembro in nueva_clase.__members__.values():
            if miembro.value == 0 and miembro.name != '_invalido':
                raise ValueError(f"{nombre} no puede tener otro miembro con valor 0 (reservado para '_invalido').")
        
        return nueva_clase

class EnumSQL(Enum,metaclass=EnumSQLMeta):
    """
    Clase base para enumeraciones que se lleva bien con SQL.
    """
    _invalido = 0  # Siempre debe ser 0 y no debe haber otro miembro con este valor.

    def __init__(self, *args, **kwds):
        pass

    @classmethod
    def desdeCadena(cls, cadena: str) -> 'EnumSQL':
        return cls.__members__.get(cadena, cls._invalido)

    def haciaCadena(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.haciaCadena()

    def __repr__(self) -> str:
        return self.haciaCadena()



from setuptools import setup, find_packages

setup(
    name='chastack-bdd',
    description="setuptools.build_meta",
    version='0.2.0',
    author           = 'Hernán A. Teszkiewicz Novick',
    author_email     = 'herni@cajadeideas.ar',
    url= 'https://github.com/Hernanatn/bdd.py',
    packages=['bdd', 'bdd.errores','bdd.tipos'],
)